package cc.smtweb.framework.core.mvc.controller.scan;

import lombok.extern.slf4j.Slf4j;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Modifier;
import java.net.JarURLConnection;
import java.net.URL;
import java.net.URLDecoder;
import java.util.Enumeration;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;

@Slf4j
public class PackageScanner {
    private ClassParser classParser;

    public PackageScanner(ClassParser classParser) {
        this.classParser = classParser;
    }

    /**
     * 1. 遍历所有目录 2. 处理注解类，进行适配类匹配 3. 放入路径映射map
     *
     * @throws IOException
     */
    public void scan(String packagePath) throws IOException {
//    log.info("[smt]start scan " + packagePath);
        directorize(packagePath, true);
//    log.info("[smt]end scan " + packagePath);
    }

    /**
     * 扫描目录下的注解类
     *
     * @param packageName 目录对象
     * @param recursive   是否递归扫描包
     */
    private void directorize(String packageName, boolean recursive)
            throws IOException {
        String packageDirName = packageName.replace('.', '/');

        Enumeration<URL> dirs = Thread.currentThread().getContextClassLoader()
                .getResources(packageDirName);

        // 循环迭代下去
        while (dirs.hasMoreElements()) {
            // 获取下一个元素
            URL url = dirs.nextElement();
            // 得到协议的名称
            String protocol = url.getProtocol();
            // 如果是以文件的形式保存在服务器上
            if ("file".equals(protocol)) {
//        log.debug("scan action in file");
                // 获取包的物理路径
                String filePath = URLDecoder.decode(url.getFile(), "UTF-8");
                // 以文件的方式扫描整个包下的文件 并添加到集合中
                findAndAddClassesInPackageByFile(packageName, "", filePath,
                        recursive);
                // 执行属性设置
//        for (FieldAutowriter field : fieldList) {
//          boolean ret = field.writeValue(this.context);
//          if (ret) {
//            log.debug("write bean " + field.getName()
//                    + " success.");
//          } else {
//            log.error("write bean " + field.getName()
//                    + " failure.");
//          }
//        }
            } else if ("jar".equals(protocol)) {
                // 如果是jar包文件
                // 定义一个JarFile
                try {
                    // 获取jar
                    JarFile jar = ((JarURLConnection) url.openConnection())
                            .getJarFile();
                    // 从此jar包 得到一个枚举类
                    Enumeration<JarEntry> entries = jar.entries();
                    // 同样的进行循环迭代
                    while (entries.hasMoreElements()) {
                        // 获取jar里的一个实体 可以是目录 和一些jar包里的其他文件 如META-INF等文件
                        JarEntry entry = entries.nextElement();
                        String name = entry.getName();
                        // 如果是以/开头的
                        if (name.charAt(0) == '/') {
                            // 获取后面的字符串
                            name = name.substring(1);
                        }
                        // 如果前半部分和定义的包名相同
                        if (name.startsWith(packageDirName)) {
                            int idx = name.lastIndexOf('/');
                            // 如果以"/"结尾 是一个包
                            if (idx != -1) {
                                // 获取包名 把"/"替换成"."
                                packageName = name.substring(0, idx).replace(
                                        '/', '.');
                            }
                            // 如果可以迭代下去 并且是一个包
                            if ((idx != -1) || recursive) {
                                // 如果是一个.class文件 而且不是目录
                                if (!entry.isDirectory()) {
                                    parseFile(
                                            packageName, "",
                                            name.substring(packageName.length() + 1));
                                }
                            }
                        }
                    }
                } catch (IOException e) {
                    log.error("scan jar error", e);
                    // log.error("在扫描用户定义视图时从jar包获取文件出错");
                    // e.printStackTrace();
                }
            }
        }
    }

    /**
     * 以文件的形式来获取包下的所有Class
     *
     * @param packageName
     * @param packagePath
     * @param recursive
     */
    public void findAndAddClassesInPackageByFile(String packageName, String dirName,
                                                 String packagePath, final boolean recursive) {
        // 获取此包的目录 建立一个File
        File dir = new File(packagePath);
        // 如果不存在或者 也不是目录就直接返回
        if (!dir.exists() || !dir.isDirectory()) {
            // log.warn("用户定义包名 " + packageName + " 下没有任何文件");
            return;
        }

        // 如果存在 就获取包下的所有文件 包括目录
        // 自定义过滤规则 如果可以循环(包含子目录) 或则是以.class结尾的文件(编译好的java类文件)
        File[] dirFiles = dir.listFiles(file -> {
            String fileName = file.getName();
            return (recursive && file.isDirectory())
                    || fileName.endsWith(".class");
        });

        // 循环所有文件
        if (dirFiles != null) {
            for (File file : dirFiles) {
                // 如果是目录 则继续扫描
                if (file.isDirectory()) {
                    findAndAddClassesInPackageByFile(packageName + "." + file.getName(),
                            file.getName(), file.getAbsolutePath(), recursive);
                } else {
                    parseFile(packageName, dirName, file.getName());
                }
            }
        }
    }

    private void parseFile(String packageName, String dirName, String fileName) {
        if (fileName.endsWith(".class")) {
            // 去掉后面的".class" 获取真正的类名
            String className = fileName.substring(0, fileName.length() - 6);

            className = packageName + '.' + className;

            try {
                Class<?> clazz = Class.forName(className);

                int modifiers = clazz.getModifiers();

                if (Modifier.isPublic(modifiers) && !Modifier.isAbstract(modifiers)) {
                    classParser.parse(dirName, clazz);
                }
            } catch (Exception e) {
                log.error("init class error " + className, e);
                throw new RuntimeException(e);
            }
        }
    }
}
